/* $Id: et4000_io.inc,v 1.6 1998/11/24 18:28:09 marcus Exp $
******************************************************************************

   Tseng ET4000 I/O

   Copyright (C) 1997      Jason McMullan, [jmcc@ggi-project.org]
   Copyright (C) 1997	     Andreas Vogler, [duff@allgaeu.org]

   Permission is hereby granted, free of charge, to any person obtaining a
   copy of this software and associated documentation files (the "Software"),
   to deal in the Software without restriction, including without limitation
   the rights to use, copy, modify, merge, publish, distribute, sublicense,
   and/or sell copies of the Software, and to permit persons to whom the
   Software is furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included in
   all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
   THE AUTHOR(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
   IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
   CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

******************************************************************************

	This code was derived from the following source of information:
	[1]     IBM vga driver by Jason McMullan
	[2]     vgadocs4 by Finn Thoegersen

******************************************************************************
*/

#include "IBM/vga_io.inc"

/*
 generalized DAC output functions
*/

static kgiu16 dac_port[4] ={ 0x3C8, 0x3C9, 0x3C6, 0x3C7 };

void DAC_out8(struct kgi_display *dpy,kgiu8 value, kgiu8 dac_reg)
{
    if (dac_reg & ~0x3)
      printk("%s: warning: DAC register select lines truncated: %x to %x.\n",\
	     __FILE__, dac_reg, dac_reg &= 0x3);
    
    io_out8(value, dac_port[dac_reg & 0x03]);
}

kgiu8 DAC_in8(struct kgi_display *dpy,kgiu8 dac_reg)
{
    if(dac_reg & ~0x03) 
      printk("%s: warning: DAC register select lines truncated: %x to %x.\n",\
	     __FILE__, dac_reg, dac_reg &= 0x03);
    
    return io_in8(dac_port[dac_reg & 0x03]);
}

void DAC_ins8(struct kgi_display *dpy,kgiu8 dac_reg,void *buf, kgi_uint count)
{
    if(dac_reg & ~0x03)
      printk("%s: warning; DAC register select lines truncated: %x to %x.\n",\
	     __FILE__, dac_reg, dac_reg &= 0x3);
    
    io_ins8(dac_port[dac_reg & 0x03], buf, count);
}

void DAC_outs8(struct kgi_display *dpy,kgiu8 dac_reg, void *buf, kgi_uint count)
{
    if(dac_reg & ~0x03) 
      printk("%s: warning: DAC register select lines truncated: %x to %x.\n",\
	     __FILE__, dac_reg, dac_reg &= 0x03);
    
    io_outs8(dac_port[dac_reg & 0x03], buf, count);
}

/*
	clock I/O functions
*/

void CLK_out8(struct kgi_display *dpy,kgiu8 value,kgiu8 reg)
{
    if (reg!=0)
      INTERNAL_ERROR;
    
    /* !!! there are 2 additional bits in 0x3d4 index 0x31 bit 6-7 */
    /* !!! but these may conflict with the DAC's RS2 */
    
    PASSERT(!(value & ~0x07), "trying to select clock: %x", value);
    
    outbMISC((io_in8(MISCr) & ~MISC_CLOCK_MASK) | ((value & 0x03) << 2));
    outbCRT((inbCRT(CRT_CCR) & ~CR34_CLK_SEL2) | ((value & 0x04) >> 1),
	    CRT_CCR);
}

kgiu8 CLK_in8(struct kgi_display *dpy,kgiu8 reg)
{
    kgiu8 ret=0;
    if (reg!=0)
      INTERNAL_ERROR;
    
    /* !!! there are 2 additional bits in 0x3d4 index 0x31 bit 6-7 */
    /* !!! but these may conflict with the DAC's RS2 */
    
    ret |= (io_in8(MISCr) & MISC_CLOCK_MASK) >> 2;
    ret |= (inbCRT(CRT_CCR) & CR34_CLK_SEL2) << 1;
    
    return ret;
}

/* I/O functions for the Sprite register */

static inline kgiu8 Sprite_in8(kgiu8 reg)
{
	io_out8(reg, SPRITEI);
	return io_in8(SPRITED);
}

static inline void Sprite_out8(kgiu8 val, kgiu8 reg)
{
	io_out16((val << 8) | reg, SPRITEI);
}

#define NoSEQRegs	0x08
#define	NoCRTRegs	0x40
#define NoGRCRegs	0x09
#define NoATCRegs	0x18

struct statetable
{
    kgiu8 MISC;
    kgiu8 SEQ[NoSEQRegs];
    kgiu8 CRT[NoCRTRegs];
    kgiu8 GRC[NoGRCRegs];
    kgiu8 ATC[NoATCRegs];
};

static void save_SEQ(kgiu8 *buf)
{
    int i;
    
    for(i = 0; i < 5; i++) 
      *(buf++)=inbSEQ(i);
    buf++;
    for(i = 6; i < 8; i++) 
      *(buf++)=inbSEQ(i);
}

static void restore_SEQ(kgiu8 *buf)
{
    register kgiu8 i;
    for(i=0; i < 5; i++) 
      outbSEQ(*(buf++),i);
    buf++;
    for(i=6; i < 8; i++) 
      outbSEQ(*(buf++),i);
}

static void save_CRT(kgiu8 *buf)
{
    int i;
    
    for(i = 0; i < 0x19; i++) 
      *(buf++)=inbCRT(i);
    buf+=0x17;
    for(i = 0x30; i < 0x40; i++) 
      *(buf++)=inbCRT(i);
}

static void restore_CRT(kgiu8 *buf)
{
    register kgiu8 i;
    
    outbCRT(inbCRT(CRT_VSYNCEND) & ~CR11_LOCKTIMING, CRT_VSYNCEND);
    outbCRT(inbCRT(CRT_HBLANKEND) | CR03_IS_VGA,CRT_HBLANKEND);
    
    for(i = 0; i < 0x19; i++) {
	switch (i) {
	case CRT_HBLANKEND:
	    *buf |= CR03_IS_VGA;
	    break;
	case CRT_VSYNCEND:
	    *buf &= ~CR11_LOCKTIMING;
	    break;
	default:
	    break;
	}
	outbCRT(*(buf++), i);
    }
    buf+=0x17;
    for(i = 0x30; i < 0x40; i++) {
	switch(i) {
	case CRT_GENERAL:
	    *buf = (io_in8(CRT_GENERAL) & ~CR31_CLK_MASK)
	         | (*buf & CR31_CLK_MASK);
	    break;
	case CRT_CCR:
	    *buf = (io_in8(CRT_CCR) & CR34_CLK_SEL2)
	         | (*buf & ~CR34_CLK_SEL2);
	    break;
	default:
	    break;
	}
	outbCRT(*(buf++), i);
    }
}

static void save_GRC(kgiu8 *buf)
{
    int i;
    
    for(i = 0; i < NoGRCRegs; i++) 
      *(buf++)=inbGRC(i);
}

static void restore_GRC(kgiu8 *buf)
{
    register kgiu8 i;
    
    for(i = 0; i < NoGRCRegs; i++) 
      outbGRC(*(buf++),i);
}

static void save_ATC(kgiu8 *buf)
{
    int i;
    
    for(i = 0; i < 0x15; i++) {
	io_in8(INS1r);  /* reset AFF and setup index */
	*(buf++)=inbATC(i);
    }
    buf++;
    for(i = 0x16; i < 0x18; i++) {
	io_in8(INS1r);  
	*(buf++)=inbATC(i);
    }
    io_in8(INS1r); 
    io_out8( ATCI_ENABLEDISPLAY, ATCI);
}

static void restore_ATC(kgiu8 *buf)
{
    register kgiu8 i;
    
    for(i = 0; i < 0x15; i++) {
	io_in8(INS1r); 
	outbATC(*(buf++),i);
    }
    buf++;
    for(i = 0x16; i < 0x18; i++) {
	io_in8(INS1r); 
	outbATC(*(buf++),i);
    }
    io_in8(INS1r); 
    io_out8(ATCI_ENABLEDISPLAY, ATCI);
}

static void
save_state(struct statetable *state)
{
    state->MISC  = io_in8(MISCr);	/* save miscellaneous output reg */
    
    save_SEQ(state->SEQ);	/* save sequencer regs */
    save_CRT(state->CRT);	/* save cathode ray tube controller regs */
    save_GRC(state->GRC);	/* save graphics controller regs */
    save_ATC(state->ATC);	/* save attribute controller regs */
}

static void restore_state(struct statetable *state)
{
    outbMISC((io_in8(MISCr) & MISC_CLOCK_MASK) | state->MISC);
    
    restore_SEQ(state->SEQ);
    restore_CRT(state->CRT);
    restore_GRC(state->GRC);
    restore_ATC(state->ATC);
}

#if defined(DEBUG_IO)||defined(DEBUG_STATE)
static void dump_state(struct statetable *state)
{ int i;
  printk("MISC: %.2x",state->MISC);
  
  for(i = 0; i < NoSEQRegs; i++)
   { if(!(i & 0xF)) printk("\nSEQ:%.2x: ",i); printk("%.2x, ",state->SEQ[i]); }
  for(i = 0; i < NoCRTRegs; i++)
   { if(!(i & 0xf)) printk("\nCRT:%.2x: ",i); printk("%.2x, ",state->CRT[i]); }
  for(i = 0; i < NoGRCRegs; i++)
   { if(!(i & 0xF)) printk("\nGRC:%.2x: ",i); printk("%.2x, ",state->GRC[i]); }
  for(i = 0; i < NoATCRegs; i++)
   { if(!(i & 0xF)) printk("\nATC:%.2x: ",i); printk("%.2x, ",state->ATC[i]); }
  printk("\n");
}
#endif

