/* 
 *  Extended crossbox (supports drawing ops)
 *
 *  Copyright (C) 1999 Nicolai Haehnle
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include "lib.h"
#include "ggi2dutil.h"

#define COPYPIXEL() \
do {								\
	getPixel(src, tx, sy, &pixel);				\
	if (pixel != last_src) {				\
		unmapPixel(src, pixel, &color);			\
		last_src = pixel;				\
		last_dst = mapColor(dest, &color);		\
	}							\
	putPixel(dest, tnx, dy, last_dst);			\
} while(0)
		

void	GGIEXP(crossBlit)(struct ggi_visual *src, ggi_sint sx, ggi_sint sy,
		ggi_sint w, ggi_sint h, struct ggi_visual *dest,
		ggi_sint dx, ggi_sint dy)
{
	void (*putPixel)(struct ggi_visual *vis, ggi_sint x, ggi_sint y, 
			ggi_pixel color);
	ggifunc_getpixel *getPixel;
	ggifunc_mapcolor *mapColor;
	ggifunc_unmappixel *unmapPixel;
	
	int tx, tnx, tw;
	ggi_pixel last_src, last_dst = -2;
	ggi_pixel pixel;
	ggi_color color;

	CLIPCOPYBOX(src,dest,sx,sy,dx,dy,w,h);
	
	/* Get some functions for faster access */
	putPixel = GGI2D_FUNC(dest)->putPixel;
	getPixel = src->opdraw->getpixel;
	mapColor = dest->opcolor->mapcolor;
	unmapPixel = src->opcolor->unmappixel;
	
	/* Safe init */
	getPixel(src, sx, sy, &last_src);
	last_src++;
	
	if (dy < sy || ((dy == sy) && (dx < sx))) {
		for(; h; h--, sy++, dy++) {
			for(tx = sx, tnx = dx, tw = w; tw; tw--, tx++, tnx++) {
				COPYPIXEL();
			}
		}
	} else if (dy > sy) {
		sy += h-1;
		dy += h-1;
		for(; h; h--, sy--, dy--) {
			for(tx = sx, tnx = dx, tw = w; tw; tw--, tx++, tnx++) {
				COPYPIXEL();
			}
		}
	} else {
		for(; h; h--, sy++, dy++) {
			for(tx = sx+w-1, tnx = dx+w-1, tw = w; tw; 
					tw--, tx--, tnx--) {
				COPYPIXEL();
			}
		}
	}
}
